/////////////////////////////////////////////////////////////
// Delegator.cpp - Generic Delegator Component
//
// Copyright 1997, Keith Brown
//
// The Delegator class holds a single interface pointer and
// implements the delegating unknown, as well as the auto-
// forwarding of method calls to the inner object.
// Delegator objects notify their CoDelegator parent when
// their individual refcount drops to zero so the parent
// can destroy them and remove them from the collection.
/////////////////////////////////////////////////////////////
#include "stdafx.h"
#include "Delegator.h"
#include "CoDelegator.h"
#include "DelegatorTLSData.h"
#include "StackTraceAccess.h"

ITypeLib* Delegator::GetObjectTypeLib()
{
	if ( m_parent.m_pTypeLib )
		return m_parent.m_pTypeLib;

	CoDelegator::Lock lock( m_parent );
	if ( !m_parent.m_pTypeLib && !m_bCheckedRegistryForTypeLib )
	{
		m_parent._loadTypeLibFromRegistry( m_iid );
		m_bCheckedRegistryForTypeLib = true;
	}
	return m_parent.m_pTypeLib;
}


// delegating unknown
STDMETHODIMP Delegator_QueryInterface( Delegator* pThis, REFIID iid, void** ppv )
{
	return pThis->m_parent.GetOuter()->QueryInterface( iid, ppv );
}

STDMETHODIMP_(ULONG) Delegator_AddRef( Delegator* pThis )
{
	// thread-safe - only the initial thread hits both resources
	if ( 0 == pThis->m_cRefs )
		pThis->m_parent.GetOuter()->AddRef();
	return InterlockedIncrement( &pThis->m_cRefs );
}

STDMETHODIMP_(ULONG) Delegator_Release( Delegator* pThis )
{
	const long cRefs = InterlockedDecrement( &pThis->m_cRefs );
	if ( 0 != cRefs )
		return cRefs;

	// thread-safe - only the final thread hits both resources
	IUnknown* pOuter = pThis->m_parent.GetOuter();
	pThis->m_parent.DeleteDelegator( pThis );
	pOuter->Release(); // once we're stabilized, release the outer.
	return 0;
}

#pragma code_seg(".orpc")

extern DWORD g_nDelegatorTLSIndex;

extern "C" {

struct PreCallParams
{
	// ORDER_DEPENDENCY (all elements)
	DWORD		vtblIndex;
	DWORD		retaddr;
	Delegator*	pThis;
	DWORD		ArgStart;	// pArgs = &ArgStart (in case we ever decide to trace
							//                    arguments to method calls)
};

static HRESULT __stdcall PreCall( const PreCallParams* pParams, DWORD ecx )
{
	long nTime = 0;
	HRESULT hr = StackTracePreCall( &nTime );
	if ( FAILED( hr ) )
		return hr;

	// push another entry onto the local stack
    DelegatorTLSData* pTail = reinterpret_cast<DelegatorTLSData*>( TlsGetValue( g_nDelegatorTLSIndex ) );
    DelegatorTLSData* pHead = new DelegatorTLSData( pParams->pThis, pParams->retaddr, ecx, nTime, pParams->vtblIndex / 4, pTail );
	if ( !pHead )
		return E_OUTOFMEMORY;
	
	TlsSetValue( g_nDelegatorTLSIndex, pHead );

	return hr;
}

struct PostCallVars
{
	// ORDER_DEPENDENCY (all elements)
	DWORD ecx;
	DWORD retval;
	DWORD retaddr;
};

static void __stdcall PostCall( PostCallVars* pvars )
{
    DelegatorTLSData* pHead = reinterpret_cast<DelegatorTLSData*>( TlsGetValue( g_nDelegatorTLSIndex ) );
	TlsSetValue( g_nDelegatorTLSIndex, pHead->m_pTail );
	
	// PreCall guarantees that pHead will be valid.

	Delegator& delegator = *pHead->m_pDelegator;
	CoDelegator& parent = delegator.m_parent;

	// recall stored value of ecx register
	pvars->ecx = pHead->m_ecx;

	// by default, just dump the index of the method
	wchar_t szMethodName[1024];
	wsprintfW( szMethodName, L"method %d", pHead->m_nMethodIndex );

	// lookup the function name instead, if we have a typelib
	ITypeLib* ptlb = delegator.GetObjectTypeLib();
	if ( ptlb )
	{
		ITypeInfo* pti = 0;
		HRESULT hr = ptlb->GetTypeInfoOfGuid( delegator.m_iid, &pti );
		if ( SUCCEEDED( hr ) )
		{
			// type libraries are so incredibly gross to deal with...
			// just had to say that :)
			TYPEATTR* pta = 0;
			hr = pti->GetTypeAttr( &pta );
			if ( SUCCEEDED( hr ) )
			{
				UINT nMethodIndex = 0;
				switch ( pta->typekind )
				{
					case TKIND_INTERFACE:	// custom interfaces
						// method indeces start just beyond IUnknown
						nMethodIndex = pHead->m_nMethodIndex - 3;
						break;
					case TKIND_DISPATCH:	// duals, dispinterfaces
						// method indeces start at top of vtbl
						nMethodIndex = pHead->m_nMethodIndex;
						break;
					default:
						// should never get here
						_ASSERTE( FALSE && "TypeInfo uses unsupported TYPEKIND" );
						break;
				}
				
				// based on the index of the method, look up the logical id
				FUNCDESC* pfd = 0;
				hr = pti->GetFuncDesc( nMethodIndex, &pfd );
				if ( SUCCEEDED( hr ) )
				{
					BSTR bstrName = 0;
					hr = pti->GetDocumentation( pfd->memid, &bstrName, 0, 0, 0 );
					if ( SUCCEEDED( hr ) )
					{
						// a bit of paranoia here
						UINT cch = SysStringLen( bstrName ) + 1;
						if ( cch < sizeof szMethodName / sizeof *szMethodName )
							CopyMemory( szMethodName, bstrName, cch * sizeof *szMethodName );
					}
					pti->ReleaseFuncDesc( pfd );
				}
				pti->ReleaseTypeAttr( pta );
			}
			pti->Release();
		}
	}

	// best effort delivery
	StackTracePostCall( pHead->m_nTime,
						parent.GetObjectName(),
						parent.GetObjectCLSID(),
						delegator.m_iid,
						szMethodName,
						pvars->retval );

	pvars->retaddr = pHead->m_retAddr;
	delete pHead;
}

}

__declspec(naked) void thunk(void)
{
	__asm
	{
		// stack diagram:
		//     0        4      8   12
		// vtblIndex/retaddr/this/args
		push ecx
		mov  ecx, esp
		lea  eax, [esp+4]
		push eax
		call PreCall			// PreCall( PreCallParams*, ecx );

		test eax, 0x80000000
		jz   PreCallSucceeded
		push 0					// note failure
		jmp  Label1
PreCallSucceeded:
		push 1                  // note success
Label1:

		// stack diagram:
		//     0             4        8     12   16
		// PreCallStatus/vtblIndex/retaddr/this/args
		mov  ecx, [esp+12]		// ecx = this (Delegator*)
		mov  eax, [ecx+4]		// eax = sink
		mov  [esp+12], eax		// munge the sink onto the stack

		mov  eax, [eax]			// eax = vptr
		mov  ecx, [esp+4]		// ecx = vtbl index (in bytes)
		add  eax, ecx			// eax = address of sink's virtual function

		pop  ecx				// ecx = success status of PreCall
		add  esp, 4				// remove vtblIndex from stack

		// stack diagram (note absence of any tracing goo):
		//    0      4    8
		// retaddr/this/args
		test ecx, 1
		jz   NoTrace			// if so, skip tracing for this call

		add  esp, 4				// remove original return address from stack
		call dword ptr [eax]	// delegate to the sink, and come back here
		jmp  Label2
NoTrace:
		// stack diagram (note absence of any extra goo):
		//    0      4    8
		// retaddr/this/args
		jmp  dword ptr [eax]	// delegate to the sink, but don't try
								// to return here, since PreCall failed.
								// this is the only way to correctly implement
								// the __stdcall calling convention since we
								// don't have a clue how to clean up the stack.
Label2:
		// desired stack diagram:
		//  0    4      8
		// ecx/retval/retaddr
		sub  esp, 12			// allocate a PostCallVars (the desired stack)
		mov  [esp+4], eax		// store retval from method call

		push esp
		call PostCall			// PostCall( &localVars )

		pop  ecx				// restore ecx
		pop  eax				// obtain retval
		ret						// return to original caller
	}
}

#define THUNK_ENTRY_POINT(n) \
static void __declspec(naked) tep_##n(void) \
{ __asm push (n*4) __asm jmp thunk }

// 0	QueryInterface
// 1	AddRef
// 2	Release
THUNK_ENTRY_POINT(3)
THUNK_ENTRY_POINT(4)
THUNK_ENTRY_POINT(5)
THUNK_ENTRY_POINT(6)
THUNK_ENTRY_POINT(7)
THUNK_ENTRY_POINT(8)
THUNK_ENTRY_POINT(9)
THUNK_ENTRY_POINT(10)
THUNK_ENTRY_POINT(11)
THUNK_ENTRY_POINT(12)
THUNK_ENTRY_POINT(13)
THUNK_ENTRY_POINT(14)
THUNK_ENTRY_POINT(15)
THUNK_ENTRY_POINT(16)
THUNK_ENTRY_POINT(17)
THUNK_ENTRY_POINT(18)
THUNK_ENTRY_POINT(19)
THUNK_ENTRY_POINT(20)
THUNK_ENTRY_POINT(21)
THUNK_ENTRY_POINT(22)
THUNK_ENTRY_POINT(23)
THUNK_ENTRY_POINT(24)
THUNK_ENTRY_POINT(25)
THUNK_ENTRY_POINT(26)
THUNK_ENTRY_POINT(27)
THUNK_ENTRY_POINT(28)
THUNK_ENTRY_POINT(29)
THUNK_ENTRY_POINT(30)
THUNK_ENTRY_POINT(31)
THUNK_ENTRY_POINT(32)
THUNK_ENTRY_POINT(33)
THUNK_ENTRY_POINT(34)
THUNK_ENTRY_POINT(35)
THUNK_ENTRY_POINT(36)
THUNK_ENTRY_POINT(37)
THUNK_ENTRY_POINT(38)
THUNK_ENTRY_POINT(39)
THUNK_ENTRY_POINT(40)
THUNK_ENTRY_POINT(41)
THUNK_ENTRY_POINT(42)
THUNK_ENTRY_POINT(43)
THUNK_ENTRY_POINT(44)
THUNK_ENTRY_POINT(45)
THUNK_ENTRY_POINT(46)
THUNK_ENTRY_POINT(47)
THUNK_ENTRY_POINT(48)
THUNK_ENTRY_POINT(49)
THUNK_ENTRY_POINT(50)
THUNK_ENTRY_POINT(51)
THUNK_ENTRY_POINT(52)
THUNK_ENTRY_POINT(53)
THUNK_ENTRY_POINT(54)
THUNK_ENTRY_POINT(55)
THUNK_ENTRY_POINT(56)
THUNK_ENTRY_POINT(57)
THUNK_ENTRY_POINT(58)
THUNK_ENTRY_POINT(59)
THUNK_ENTRY_POINT(60)
THUNK_ENTRY_POINT(61)
THUNK_ENTRY_POINT(62)
THUNK_ENTRY_POINT(63)
THUNK_ENTRY_POINT(64)
THUNK_ENTRY_POINT(65)
THUNK_ENTRY_POINT(66)
THUNK_ENTRY_POINT(67)
THUNK_ENTRY_POINT(68)
THUNK_ENTRY_POINT(69)
THUNK_ENTRY_POINT(70)
THUNK_ENTRY_POINT(71)
THUNK_ENTRY_POINT(72)
THUNK_ENTRY_POINT(73)
THUNK_ENTRY_POINT(74)
THUNK_ENTRY_POINT(75)
THUNK_ENTRY_POINT(76)
THUNK_ENTRY_POINT(77)
THUNK_ENTRY_POINT(78)
THUNK_ENTRY_POINT(79)
THUNK_ENTRY_POINT(80)
THUNK_ENTRY_POINT(81)
THUNK_ENTRY_POINT(82)
THUNK_ENTRY_POINT(83)
THUNK_ENTRY_POINT(84)
THUNK_ENTRY_POINT(85)
THUNK_ENTRY_POINT(86)
THUNK_ENTRY_POINT(87)
THUNK_ENTRY_POINT(88)
THUNK_ENTRY_POINT(89)
THUNK_ENTRY_POINT(90)
THUNK_ENTRY_POINT(91)
THUNK_ENTRY_POINT(92)
THUNK_ENTRY_POINT(93)
THUNK_ENTRY_POINT(94)
THUNK_ENTRY_POINT(95)
THUNK_ENTRY_POINT(96)
THUNK_ENTRY_POINT(97)
THUNK_ENTRY_POINT(98)
THUNK_ENTRY_POINT(99)
THUNK_ENTRY_POINT(100)
THUNK_ENTRY_POINT(101)
THUNK_ENTRY_POINT(102)
THUNK_ENTRY_POINT(103)
THUNK_ENTRY_POINT(104)
THUNK_ENTRY_POINT(105)
THUNK_ENTRY_POINT(106)
THUNK_ENTRY_POINT(107)
THUNK_ENTRY_POINT(108)
THUNK_ENTRY_POINT(109)
THUNK_ENTRY_POINT(110)
THUNK_ENTRY_POINT(111)
THUNK_ENTRY_POINT(112)
THUNK_ENTRY_POINT(113)
THUNK_ENTRY_POINT(114)
THUNK_ENTRY_POINT(115)
THUNK_ENTRY_POINT(116)
THUNK_ENTRY_POINT(117)
THUNK_ENTRY_POINT(118)
THUNK_ENTRY_POINT(119)
THUNK_ENTRY_POINT(120)
THUNK_ENTRY_POINT(121)
THUNK_ENTRY_POINT(122)
THUNK_ENTRY_POINT(123)
THUNK_ENTRY_POINT(124)
THUNK_ENTRY_POINT(125)
THUNK_ENTRY_POINT(126)
THUNK_ENTRY_POINT(127)
THUNK_ENTRY_POINT(128)
THUNK_ENTRY_POINT(129)
THUNK_ENTRY_POINT(130)
THUNK_ENTRY_POINT(131)
THUNK_ENTRY_POINT(132)
THUNK_ENTRY_POINT(133)
THUNK_ENTRY_POINT(134)
THUNK_ENTRY_POINT(135)
THUNK_ENTRY_POINT(136)
THUNK_ENTRY_POINT(137)
THUNK_ENTRY_POINT(138)
THUNK_ENTRY_POINT(139)
THUNK_ENTRY_POINT(140)
THUNK_ENTRY_POINT(141)
THUNK_ENTRY_POINT(142)
THUNK_ENTRY_POINT(143)
THUNK_ENTRY_POINT(144)
THUNK_ENTRY_POINT(145)
THUNK_ENTRY_POINT(146)
THUNK_ENTRY_POINT(147)
THUNK_ENTRY_POINT(148)
THUNK_ENTRY_POINT(149)
THUNK_ENTRY_POINT(150)
THUNK_ENTRY_POINT(151)
THUNK_ENTRY_POINT(152)
THUNK_ENTRY_POINT(153)
THUNK_ENTRY_POINT(154)
THUNK_ENTRY_POINT(155)
THUNK_ENTRY_POINT(156)
THUNK_ENTRY_POINT(157)
THUNK_ENTRY_POINT(158)
THUNK_ENTRY_POINT(159)
THUNK_ENTRY_POINT(160)
THUNK_ENTRY_POINT(161)
THUNK_ENTRY_POINT(162)
THUNK_ENTRY_POINT(163)
THUNK_ENTRY_POINT(164)
THUNK_ENTRY_POINT(165)
THUNK_ENTRY_POINT(166)
THUNK_ENTRY_POINT(167)
THUNK_ENTRY_POINT(168)
THUNK_ENTRY_POINT(169)
THUNK_ENTRY_POINT(170)
THUNK_ENTRY_POINT(171)
THUNK_ENTRY_POINT(172)
THUNK_ENTRY_POINT(173)
THUNK_ENTRY_POINT(174)
THUNK_ENTRY_POINT(175)
THUNK_ENTRY_POINT(176)
THUNK_ENTRY_POINT(177)
THUNK_ENTRY_POINT(178)
THUNK_ENTRY_POINT(179)
THUNK_ENTRY_POINT(180)
THUNK_ENTRY_POINT(181)
THUNK_ENTRY_POINT(182)
THUNK_ENTRY_POINT(183)
THUNK_ENTRY_POINT(184)
THUNK_ENTRY_POINT(185)
THUNK_ENTRY_POINT(186)
THUNK_ENTRY_POINT(187)
THUNK_ENTRY_POINT(188)
THUNK_ENTRY_POINT(189)
THUNK_ENTRY_POINT(190)
THUNK_ENTRY_POINT(191)
THUNK_ENTRY_POINT(192)
THUNK_ENTRY_POINT(193)
THUNK_ENTRY_POINT(194)
THUNK_ENTRY_POINT(195)
THUNK_ENTRY_POINT(196)
THUNK_ENTRY_POINT(197)
THUNK_ENTRY_POINT(198)
THUNK_ENTRY_POINT(199)
THUNK_ENTRY_POINT(200)
THUNK_ENTRY_POINT(201)
THUNK_ENTRY_POINT(202)
THUNK_ENTRY_POINT(203)
THUNK_ENTRY_POINT(204)
THUNK_ENTRY_POINT(205)
THUNK_ENTRY_POINT(206)
THUNK_ENTRY_POINT(207)
THUNK_ENTRY_POINT(208)
THUNK_ENTRY_POINT(209)
THUNK_ENTRY_POINT(210)
THUNK_ENTRY_POINT(211)
THUNK_ENTRY_POINT(212)
THUNK_ENTRY_POINT(213)
THUNK_ENTRY_POINT(214)
THUNK_ENTRY_POINT(215)
THUNK_ENTRY_POINT(216)
THUNK_ENTRY_POINT(217)
THUNK_ENTRY_POINT(218)
THUNK_ENTRY_POINT(219)
THUNK_ENTRY_POINT(220)
THUNK_ENTRY_POINT(221)
THUNK_ENTRY_POINT(222)
THUNK_ENTRY_POINT(223)
THUNK_ENTRY_POINT(224)
THUNK_ENTRY_POINT(225)
THUNK_ENTRY_POINT(226)
THUNK_ENTRY_POINT(227)
THUNK_ENTRY_POINT(228)
THUNK_ENTRY_POINT(229)
THUNK_ENTRY_POINT(230)
THUNK_ENTRY_POINT(231)
THUNK_ENTRY_POINT(232)
THUNK_ENTRY_POINT(233)
THUNK_ENTRY_POINT(234)
THUNK_ENTRY_POINT(235)
THUNK_ENTRY_POINT(236)
THUNK_ENTRY_POINT(237)
THUNK_ENTRY_POINT(238)
THUNK_ENTRY_POINT(239)
THUNK_ENTRY_POINT(240)
THUNK_ENTRY_POINT(241)
THUNK_ENTRY_POINT(242)
THUNK_ENTRY_POINT(243)
THUNK_ENTRY_POINT(244)
THUNK_ENTRY_POINT(245)
THUNK_ENTRY_POINT(246)
THUNK_ENTRY_POINT(247)
THUNK_ENTRY_POINT(248)
THUNK_ENTRY_POINT(249)
THUNK_ENTRY_POINT(250)
THUNK_ENTRY_POINT(251)
THUNK_ENTRY_POINT(252)
THUNK_ENTRY_POINT(253)
THUNK_ENTRY_POINT(254)
THUNK_ENTRY_POINT(255)
THUNK_ENTRY_POINT(256)
THUNK_ENTRY_POINT(257)
THUNK_ENTRY_POINT(258)
THUNK_ENTRY_POINT(259)
THUNK_ENTRY_POINT(260)
THUNK_ENTRY_POINT(261)
THUNK_ENTRY_POINT(262)
THUNK_ENTRY_POINT(263)
THUNK_ENTRY_POINT(264)
THUNK_ENTRY_POINT(265)
THUNK_ENTRY_POINT(266)
THUNK_ENTRY_POINT(267)
THUNK_ENTRY_POINT(268)
THUNK_ENTRY_POINT(269)
THUNK_ENTRY_POINT(270)
THUNK_ENTRY_POINT(271)
THUNK_ENTRY_POINT(272)
THUNK_ENTRY_POINT(273)
THUNK_ENTRY_POINT(274)
THUNK_ENTRY_POINT(275)
THUNK_ENTRY_POINT(276)
THUNK_ENTRY_POINT(277)
THUNK_ENTRY_POINT(278)
THUNK_ENTRY_POINT(279)
THUNK_ENTRY_POINT(280)
THUNK_ENTRY_POINT(281)
THUNK_ENTRY_POINT(282)
THUNK_ENTRY_POINT(283)
THUNK_ENTRY_POINT(284)
THUNK_ENTRY_POINT(285)
THUNK_ENTRY_POINT(286)
THUNK_ENTRY_POINT(287)
THUNK_ENTRY_POINT(288)
THUNK_ENTRY_POINT(289)
THUNK_ENTRY_POINT(290)
THUNK_ENTRY_POINT(291)
THUNK_ENTRY_POINT(292)
THUNK_ENTRY_POINT(293)
THUNK_ENTRY_POINT(294)
THUNK_ENTRY_POINT(295)
THUNK_ENTRY_POINT(296)
THUNK_ENTRY_POINT(297)
THUNK_ENTRY_POINT(298)
THUNK_ENTRY_POINT(299)
THUNK_ENTRY_POINT(300)
THUNK_ENTRY_POINT(301)
THUNK_ENTRY_POINT(302)
THUNK_ENTRY_POINT(303)
THUNK_ENTRY_POINT(304)
THUNK_ENTRY_POINT(305)
THUNK_ENTRY_POINT(306)
THUNK_ENTRY_POINT(307)
THUNK_ENTRY_POINT(308)
THUNK_ENTRY_POINT(309)
THUNK_ENTRY_POINT(310)
THUNK_ENTRY_POINT(311)
THUNK_ENTRY_POINT(312)
THUNK_ENTRY_POINT(313)
THUNK_ENTRY_POINT(314)
THUNK_ENTRY_POINT(315)
THUNK_ENTRY_POINT(316)
THUNK_ENTRY_POINT(317)
THUNK_ENTRY_POINT(318)
THUNK_ENTRY_POINT(319)
THUNK_ENTRY_POINT(320)
THUNK_ENTRY_POINT(321)
THUNK_ENTRY_POINT(322)
THUNK_ENTRY_POINT(323)
THUNK_ENTRY_POINT(324)
THUNK_ENTRY_POINT(325)
THUNK_ENTRY_POINT(326)
THUNK_ENTRY_POINT(327)
THUNK_ENTRY_POINT(328)
THUNK_ENTRY_POINT(329)
THUNK_ENTRY_POINT(330)
THUNK_ENTRY_POINT(331)
THUNK_ENTRY_POINT(332)
THUNK_ENTRY_POINT(333)
THUNK_ENTRY_POINT(334)
THUNK_ENTRY_POINT(335)
THUNK_ENTRY_POINT(336)
THUNK_ENTRY_POINT(337)
THUNK_ENTRY_POINT(338)
THUNK_ENTRY_POINT(339)
THUNK_ENTRY_POINT(340)
THUNK_ENTRY_POINT(341)
THUNK_ENTRY_POINT(342)
THUNK_ENTRY_POINT(343)
THUNK_ENTRY_POINT(344)
THUNK_ENTRY_POINT(345)
THUNK_ENTRY_POINT(346)
THUNK_ENTRY_POINT(347)
THUNK_ENTRY_POINT(348)
THUNK_ENTRY_POINT(349)
THUNK_ENTRY_POINT(350)
THUNK_ENTRY_POINT(351)
THUNK_ENTRY_POINT(352)
THUNK_ENTRY_POINT(353)
THUNK_ENTRY_POINT(354)
THUNK_ENTRY_POINT(355)
THUNK_ENTRY_POINT(356)
THUNK_ENTRY_POINT(357)
THUNK_ENTRY_POINT(358)
THUNK_ENTRY_POINT(359)
THUNK_ENTRY_POINT(360)
THUNK_ENTRY_POINT(361)
THUNK_ENTRY_POINT(362)
THUNK_ENTRY_POINT(363)
THUNK_ENTRY_POINT(364)
THUNK_ENTRY_POINT(365)
THUNK_ENTRY_POINT(366)
THUNK_ENTRY_POINT(367)
THUNK_ENTRY_POINT(368)
THUNK_ENTRY_POINT(369)
THUNK_ENTRY_POINT(370)
THUNK_ENTRY_POINT(371)
THUNK_ENTRY_POINT(372)
THUNK_ENTRY_POINT(373)
THUNK_ENTRY_POINT(374)
THUNK_ENTRY_POINT(375)
THUNK_ENTRY_POINT(376)
THUNK_ENTRY_POINT(377)
THUNK_ENTRY_POINT(378)
THUNK_ENTRY_POINT(379)
THUNK_ENTRY_POINT(380)
THUNK_ENTRY_POINT(381)
THUNK_ENTRY_POINT(382)
THUNK_ENTRY_POINT(383)
THUNK_ENTRY_POINT(384)
THUNK_ENTRY_POINT(385)
THUNK_ENTRY_POINT(386)
THUNK_ENTRY_POINT(387)
THUNK_ENTRY_POINT(388)
THUNK_ENTRY_POINT(389)
THUNK_ENTRY_POINT(390)
THUNK_ENTRY_POINT(391)
THUNK_ENTRY_POINT(392)
THUNK_ENTRY_POINT(393)
THUNK_ENTRY_POINT(394)
THUNK_ENTRY_POINT(395)
THUNK_ENTRY_POINT(396)
THUNK_ENTRY_POINT(397)
THUNK_ENTRY_POINT(398)
THUNK_ENTRY_POINT(399)
THUNK_ENTRY_POINT(400)
THUNK_ENTRY_POINT(401)
THUNK_ENTRY_POINT(402)
THUNK_ENTRY_POINT(403)
THUNK_ENTRY_POINT(404)
THUNK_ENTRY_POINT(405)
THUNK_ENTRY_POINT(406)
THUNK_ENTRY_POINT(407)
THUNK_ENTRY_POINT(408)
THUNK_ENTRY_POINT(409)
THUNK_ENTRY_POINT(410)
THUNK_ENTRY_POINT(411)
THUNK_ENTRY_POINT(412)
THUNK_ENTRY_POINT(413)
THUNK_ENTRY_POINT(414)
THUNK_ENTRY_POINT(415)
THUNK_ENTRY_POINT(416)
THUNK_ENTRY_POINT(417)
THUNK_ENTRY_POINT(418)
THUNK_ENTRY_POINT(419)
THUNK_ENTRY_POINT(420)
THUNK_ENTRY_POINT(421)
THUNK_ENTRY_POINT(422)
THUNK_ENTRY_POINT(423)
THUNK_ENTRY_POINT(424)
THUNK_ENTRY_POINT(425)
THUNK_ENTRY_POINT(426)
THUNK_ENTRY_POINT(427)
THUNK_ENTRY_POINT(428)
THUNK_ENTRY_POINT(429)
THUNK_ENTRY_POINT(430)
THUNK_ENTRY_POINT(431)
THUNK_ENTRY_POINT(432)
THUNK_ENTRY_POINT(433)
THUNK_ENTRY_POINT(434)
THUNK_ENTRY_POINT(435)
THUNK_ENTRY_POINT(436)
THUNK_ENTRY_POINT(437)
THUNK_ENTRY_POINT(438)
THUNK_ENTRY_POINT(439)
THUNK_ENTRY_POINT(440)
THUNK_ENTRY_POINT(441)
THUNK_ENTRY_POINT(442)
THUNK_ENTRY_POINT(443)
THUNK_ENTRY_POINT(444)
THUNK_ENTRY_POINT(445)
THUNK_ENTRY_POINT(446)
THUNK_ENTRY_POINT(447)
THUNK_ENTRY_POINT(448)
THUNK_ENTRY_POINT(449)
THUNK_ENTRY_POINT(450)
THUNK_ENTRY_POINT(451)
THUNK_ENTRY_POINT(452)
THUNK_ENTRY_POINT(453)
THUNK_ENTRY_POINT(454)
THUNK_ENTRY_POINT(455)
THUNK_ENTRY_POINT(456)
THUNK_ENTRY_POINT(457)
THUNK_ENTRY_POINT(458)
THUNK_ENTRY_POINT(459)
THUNK_ENTRY_POINT(460)
THUNK_ENTRY_POINT(461)
THUNK_ENTRY_POINT(462)
THUNK_ENTRY_POINT(463)
THUNK_ENTRY_POINT(464)
THUNK_ENTRY_POINT(465)
THUNK_ENTRY_POINT(466)
THUNK_ENTRY_POINT(467)
THUNK_ENTRY_POINT(468)
THUNK_ENTRY_POINT(469)
THUNK_ENTRY_POINT(470)
THUNK_ENTRY_POINT(471)
THUNK_ENTRY_POINT(472)
THUNK_ENTRY_POINT(473)
THUNK_ENTRY_POINT(474)
THUNK_ENTRY_POINT(475)
THUNK_ENTRY_POINT(476)
THUNK_ENTRY_POINT(477)
THUNK_ENTRY_POINT(478)
THUNK_ENTRY_POINT(479)
THUNK_ENTRY_POINT(480)
THUNK_ENTRY_POINT(481)
THUNK_ENTRY_POINT(482)
THUNK_ENTRY_POINT(483)
THUNK_ENTRY_POINT(484)
THUNK_ENTRY_POINT(485)
THUNK_ENTRY_POINT(486)
THUNK_ENTRY_POINT(487)
THUNK_ENTRY_POINT(488)
THUNK_ENTRY_POINT(489)
THUNK_ENTRY_POINT(490)
THUNK_ENTRY_POINT(491)
THUNK_ENTRY_POINT(492)
THUNK_ENTRY_POINT(493)
THUNK_ENTRY_POINT(494)
THUNK_ENTRY_POINT(495)
THUNK_ENTRY_POINT(496)
THUNK_ENTRY_POINT(497)
THUNK_ENTRY_POINT(498)
THUNK_ENTRY_POINT(499)
THUNK_ENTRY_POINT(500)
THUNK_ENTRY_POINT(501)
THUNK_ENTRY_POINT(502)
THUNK_ENTRY_POINT(503)
THUNK_ENTRY_POINT(504)
THUNK_ENTRY_POINT(505)
THUNK_ENTRY_POINT(506)
THUNK_ENTRY_POINT(507)
THUNK_ENTRY_POINT(508)
THUNK_ENTRY_POINT(509)
THUNK_ENTRY_POINT(510)
THUNK_ENTRY_POINT(511)
THUNK_ENTRY_POINT(512)
THUNK_ENTRY_POINT(513)
THUNK_ENTRY_POINT(514)
THUNK_ENTRY_POINT(515)
THUNK_ENTRY_POINT(516)
THUNK_ENTRY_POINT(517)
THUNK_ENTRY_POINT(518)
THUNK_ENTRY_POINT(519)
THUNK_ENTRY_POINT(520)
THUNK_ENTRY_POINT(521)
THUNK_ENTRY_POINT(522)
THUNK_ENTRY_POINT(523)
THUNK_ENTRY_POINT(524)
THUNK_ENTRY_POINT(525)
THUNK_ENTRY_POINT(526)
THUNK_ENTRY_POINT(527)
THUNK_ENTRY_POINT(528)
THUNK_ENTRY_POINT(529)
THUNK_ENTRY_POINT(530)
THUNK_ENTRY_POINT(531)
THUNK_ENTRY_POINT(532)
THUNK_ENTRY_POINT(533)
THUNK_ENTRY_POINT(534)
THUNK_ENTRY_POINT(535)
THUNK_ENTRY_POINT(536)
THUNK_ENTRY_POINT(537)
THUNK_ENTRY_POINT(538)
THUNK_ENTRY_POINT(539)
THUNK_ENTRY_POINT(540)
THUNK_ENTRY_POINT(541)
THUNK_ENTRY_POINT(542)
THUNK_ENTRY_POINT(543)
THUNK_ENTRY_POINT(544)
THUNK_ENTRY_POINT(545)
THUNK_ENTRY_POINT(546)
THUNK_ENTRY_POINT(547)
THUNK_ENTRY_POINT(548)
THUNK_ENTRY_POINT(549)
THUNK_ENTRY_POINT(550)
THUNK_ENTRY_POINT(551)
THUNK_ENTRY_POINT(552)
THUNK_ENTRY_POINT(553)
THUNK_ENTRY_POINT(554)
THUNK_ENTRY_POINT(555)
THUNK_ENTRY_POINT(556)
THUNK_ENTRY_POINT(557)
THUNK_ENTRY_POINT(558)
THUNK_ENTRY_POINT(559)
THUNK_ENTRY_POINT(560)
THUNK_ENTRY_POINT(561)
THUNK_ENTRY_POINT(562)
THUNK_ENTRY_POINT(563)
THUNK_ENTRY_POINT(564)
THUNK_ENTRY_POINT(565)
THUNK_ENTRY_POINT(566)
THUNK_ENTRY_POINT(567)
THUNK_ENTRY_POINT(568)
THUNK_ENTRY_POINT(569)
THUNK_ENTRY_POINT(570)
THUNK_ENTRY_POINT(571)
THUNK_ENTRY_POINT(572)
THUNK_ENTRY_POINT(573)
THUNK_ENTRY_POINT(574)
THUNK_ENTRY_POINT(575)
THUNK_ENTRY_POINT(576)
THUNK_ENTRY_POINT(577)
THUNK_ENTRY_POINT(578)
THUNK_ENTRY_POINT(579)
THUNK_ENTRY_POINT(580)
THUNK_ENTRY_POINT(581)
THUNK_ENTRY_POINT(582)
THUNK_ENTRY_POINT(583)
THUNK_ENTRY_POINT(584)
THUNK_ENTRY_POINT(585)
THUNK_ENTRY_POINT(586)
THUNK_ENTRY_POINT(587)
THUNK_ENTRY_POINT(588)
THUNK_ENTRY_POINT(589)
THUNK_ENTRY_POINT(590)
THUNK_ENTRY_POINT(591)
THUNK_ENTRY_POINT(592)
THUNK_ENTRY_POINT(593)
THUNK_ENTRY_POINT(594)
THUNK_ENTRY_POINT(595)
THUNK_ENTRY_POINT(596)
THUNK_ENTRY_POINT(597)
THUNK_ENTRY_POINT(598)
THUNK_ENTRY_POINT(599)
THUNK_ENTRY_POINT(600)
THUNK_ENTRY_POINT(601)
THUNK_ENTRY_POINT(602)
THUNK_ENTRY_POINT(603)
THUNK_ENTRY_POINT(604)
THUNK_ENTRY_POINT(605)
THUNK_ENTRY_POINT(606)
THUNK_ENTRY_POINT(607)
THUNK_ENTRY_POINT(608)
THUNK_ENTRY_POINT(609)
THUNK_ENTRY_POINT(610)
THUNK_ENTRY_POINT(611)
THUNK_ENTRY_POINT(612)
THUNK_ENTRY_POINT(613)
THUNK_ENTRY_POINT(614)
THUNK_ENTRY_POINT(615)
THUNK_ENTRY_POINT(616)
THUNK_ENTRY_POINT(617)
THUNK_ENTRY_POINT(618)
THUNK_ENTRY_POINT(619)
THUNK_ENTRY_POINT(620)
THUNK_ENTRY_POINT(621)
THUNK_ENTRY_POINT(622)
THUNK_ENTRY_POINT(623)
THUNK_ENTRY_POINT(624)
THUNK_ENTRY_POINT(625)
THUNK_ENTRY_POINT(626)
THUNK_ENTRY_POINT(627)
THUNK_ENTRY_POINT(628)
THUNK_ENTRY_POINT(629)
THUNK_ENTRY_POINT(630)
THUNK_ENTRY_POINT(631)
THUNK_ENTRY_POINT(632)
THUNK_ENTRY_POINT(633)
THUNK_ENTRY_POINT(634)
THUNK_ENTRY_POINT(635)
THUNK_ENTRY_POINT(636)
THUNK_ENTRY_POINT(637)
THUNK_ENTRY_POINT(638)
THUNK_ENTRY_POINT(639)
THUNK_ENTRY_POINT(640)
THUNK_ENTRY_POINT(641)
THUNK_ENTRY_POINT(642)
THUNK_ENTRY_POINT(643)
THUNK_ENTRY_POINT(644)
THUNK_ENTRY_POINT(645)
THUNK_ENTRY_POINT(646)
THUNK_ENTRY_POINT(647)
THUNK_ENTRY_POINT(648)
THUNK_ENTRY_POINT(649)
THUNK_ENTRY_POINT(650)
THUNK_ENTRY_POINT(651)
THUNK_ENTRY_POINT(652)
THUNK_ENTRY_POINT(653)
THUNK_ENTRY_POINT(654)
THUNK_ENTRY_POINT(655)
THUNK_ENTRY_POINT(656)
THUNK_ENTRY_POINT(657)
THUNK_ENTRY_POINT(658)
THUNK_ENTRY_POINT(659)
THUNK_ENTRY_POINT(660)
THUNK_ENTRY_POINT(661)
THUNK_ENTRY_POINT(662)
THUNK_ENTRY_POINT(663)
THUNK_ENTRY_POINT(664)
THUNK_ENTRY_POINT(665)
THUNK_ENTRY_POINT(666)
THUNK_ENTRY_POINT(667)
THUNK_ENTRY_POINT(668)
THUNK_ENTRY_POINT(669)
THUNK_ENTRY_POINT(670)
THUNK_ENTRY_POINT(671)
THUNK_ENTRY_POINT(672)
THUNK_ENTRY_POINT(673)
THUNK_ENTRY_POINT(674)
THUNK_ENTRY_POINT(675)
THUNK_ENTRY_POINT(676)
THUNK_ENTRY_POINT(677)
THUNK_ENTRY_POINT(678)
THUNK_ENTRY_POINT(679)
THUNK_ENTRY_POINT(680)
THUNK_ENTRY_POINT(681)
THUNK_ENTRY_POINT(682)
THUNK_ENTRY_POINT(683)
THUNK_ENTRY_POINT(684)
THUNK_ENTRY_POINT(685)
THUNK_ENTRY_POINT(686)
THUNK_ENTRY_POINT(687)
THUNK_ENTRY_POINT(688)
THUNK_ENTRY_POINT(689)
THUNK_ENTRY_POINT(690)
THUNK_ENTRY_POINT(691)
THUNK_ENTRY_POINT(692)
THUNK_ENTRY_POINT(693)
THUNK_ENTRY_POINT(694)
THUNK_ENTRY_POINT(695)
THUNK_ENTRY_POINT(696)
THUNK_ENTRY_POINT(697)
THUNK_ENTRY_POINT(698)
THUNK_ENTRY_POINT(699)
THUNK_ENTRY_POINT(700)
THUNK_ENTRY_POINT(701)
THUNK_ENTRY_POINT(702)
THUNK_ENTRY_POINT(703)
THUNK_ENTRY_POINT(704)
THUNK_ENTRY_POINT(705)
THUNK_ENTRY_POINT(706)
THUNK_ENTRY_POINT(707)
THUNK_ENTRY_POINT(708)
THUNK_ENTRY_POINT(709)
THUNK_ENTRY_POINT(710)
THUNK_ENTRY_POINT(711)
THUNK_ENTRY_POINT(712)
THUNK_ENTRY_POINT(713)
THUNK_ENTRY_POINT(714)
THUNK_ENTRY_POINT(715)
THUNK_ENTRY_POINT(716)
THUNK_ENTRY_POINT(717)
THUNK_ENTRY_POINT(718)
THUNK_ENTRY_POINT(719)
THUNK_ENTRY_POINT(720)
THUNK_ENTRY_POINT(721)
THUNK_ENTRY_POINT(722)
THUNK_ENTRY_POINT(723)
THUNK_ENTRY_POINT(724)
THUNK_ENTRY_POINT(725)
THUNK_ENTRY_POINT(726)
THUNK_ENTRY_POINT(727)
THUNK_ENTRY_POINT(728)
THUNK_ENTRY_POINT(729)
THUNK_ENTRY_POINT(730)
THUNK_ENTRY_POINT(731)
THUNK_ENTRY_POINT(732)
THUNK_ENTRY_POINT(733)
THUNK_ENTRY_POINT(734)
THUNK_ENTRY_POINT(735)
THUNK_ENTRY_POINT(736)
THUNK_ENTRY_POINT(737)
THUNK_ENTRY_POINT(738)
THUNK_ENTRY_POINT(739)
THUNK_ENTRY_POINT(740)
THUNK_ENTRY_POINT(741)
THUNK_ENTRY_POINT(742)
THUNK_ENTRY_POINT(743)
THUNK_ENTRY_POINT(744)
THUNK_ENTRY_POINT(745)
THUNK_ENTRY_POINT(746)
THUNK_ENTRY_POINT(747)
THUNK_ENTRY_POINT(748)
THUNK_ENTRY_POINT(749)
THUNK_ENTRY_POINT(750)
THUNK_ENTRY_POINT(751)
THUNK_ENTRY_POINT(752)
THUNK_ENTRY_POINT(753)
THUNK_ENTRY_POINT(754)
THUNK_ENTRY_POINT(755)
THUNK_ENTRY_POINT(756)
THUNK_ENTRY_POINT(757)
THUNK_ENTRY_POINT(758)
THUNK_ENTRY_POINT(759)
THUNK_ENTRY_POINT(760)
THUNK_ENTRY_POINT(761)
THUNK_ENTRY_POINT(762)
THUNK_ENTRY_POINT(763)
THUNK_ENTRY_POINT(764)
THUNK_ENTRY_POINT(765)
THUNK_ENTRY_POINT(766)
THUNK_ENTRY_POINT(767)
THUNK_ENTRY_POINT(768)
THUNK_ENTRY_POINT(769)
THUNK_ENTRY_POINT(770)
THUNK_ENTRY_POINT(771)
THUNK_ENTRY_POINT(772)
THUNK_ENTRY_POINT(773)
THUNK_ENTRY_POINT(774)
THUNK_ENTRY_POINT(775)
THUNK_ENTRY_POINT(776)
THUNK_ENTRY_POINT(777)
THUNK_ENTRY_POINT(778)
THUNK_ENTRY_POINT(779)
THUNK_ENTRY_POINT(780)
THUNK_ENTRY_POINT(781)
THUNK_ENTRY_POINT(782)
THUNK_ENTRY_POINT(783)
THUNK_ENTRY_POINT(784)
THUNK_ENTRY_POINT(785)
THUNK_ENTRY_POINT(786)
THUNK_ENTRY_POINT(787)
THUNK_ENTRY_POINT(788)
THUNK_ENTRY_POINT(789)
THUNK_ENTRY_POINT(790)
THUNK_ENTRY_POINT(791)
THUNK_ENTRY_POINT(792)
THUNK_ENTRY_POINT(793)
THUNK_ENTRY_POINT(794)
THUNK_ENTRY_POINT(795)
THUNK_ENTRY_POINT(796)
THUNK_ENTRY_POINT(797)
THUNK_ENTRY_POINT(798)
THUNK_ENTRY_POINT(799)
THUNK_ENTRY_POINT(800)
THUNK_ENTRY_POINT(801)
THUNK_ENTRY_POINT(802)
THUNK_ENTRY_POINT(803)
THUNK_ENTRY_POINT(804)
THUNK_ENTRY_POINT(805)
THUNK_ENTRY_POINT(806)
THUNK_ENTRY_POINT(807)
THUNK_ENTRY_POINT(808)
THUNK_ENTRY_POINT(809)
THUNK_ENTRY_POINT(810)
THUNK_ENTRY_POINT(811)
THUNK_ENTRY_POINT(812)
THUNK_ENTRY_POINT(813)
THUNK_ENTRY_POINT(814)
THUNK_ENTRY_POINT(815)
THUNK_ENTRY_POINT(816)
THUNK_ENTRY_POINT(817)
THUNK_ENTRY_POINT(818)
THUNK_ENTRY_POINT(819)
THUNK_ENTRY_POINT(820)
THUNK_ENTRY_POINT(821)
THUNK_ENTRY_POINT(822)
THUNK_ENTRY_POINT(823)
THUNK_ENTRY_POINT(824)
THUNK_ENTRY_POINT(825)
THUNK_ENTRY_POINT(826)
THUNK_ENTRY_POINT(827)
THUNK_ENTRY_POINT(828)
THUNK_ENTRY_POINT(829)
THUNK_ENTRY_POINT(830)
THUNK_ENTRY_POINT(831)
THUNK_ENTRY_POINT(832)
THUNK_ENTRY_POINT(833)
THUNK_ENTRY_POINT(834)
THUNK_ENTRY_POINT(835)
THUNK_ENTRY_POINT(836)
THUNK_ENTRY_POINT(837)
THUNK_ENTRY_POINT(838)
THUNK_ENTRY_POINT(839)
THUNK_ENTRY_POINT(840)
THUNK_ENTRY_POINT(841)
THUNK_ENTRY_POINT(842)
THUNK_ENTRY_POINT(843)
THUNK_ENTRY_POINT(844)
THUNK_ENTRY_POINT(845)
THUNK_ENTRY_POINT(846)
THUNK_ENTRY_POINT(847)
THUNK_ENTRY_POINT(848)
THUNK_ENTRY_POINT(849)
THUNK_ENTRY_POINT(850)
THUNK_ENTRY_POINT(851)
THUNK_ENTRY_POINT(852)
THUNK_ENTRY_POINT(853)
THUNK_ENTRY_POINT(854)
THUNK_ENTRY_POINT(855)
THUNK_ENTRY_POINT(856)
THUNK_ENTRY_POINT(857)
THUNK_ENTRY_POINT(858)
THUNK_ENTRY_POINT(859)
THUNK_ENTRY_POINT(860)
THUNK_ENTRY_POINT(861)
THUNK_ENTRY_POINT(862)
THUNK_ENTRY_POINT(863)
THUNK_ENTRY_POINT(864)
THUNK_ENTRY_POINT(865)
THUNK_ENTRY_POINT(866)
THUNK_ENTRY_POINT(867)
THUNK_ENTRY_POINT(868)
THUNK_ENTRY_POINT(869)
THUNK_ENTRY_POINT(870)
THUNK_ENTRY_POINT(871)
THUNK_ENTRY_POINT(872)
THUNK_ENTRY_POINT(873)
THUNK_ENTRY_POINT(874)
THUNK_ENTRY_POINT(875)
THUNK_ENTRY_POINT(876)
THUNK_ENTRY_POINT(877)
THUNK_ENTRY_POINT(878)
THUNK_ENTRY_POINT(879)
THUNK_ENTRY_POINT(880)
THUNK_ENTRY_POINT(881)
THUNK_ENTRY_POINT(882)
THUNK_ENTRY_POINT(883)
THUNK_ENTRY_POINT(884)
THUNK_ENTRY_POINT(885)
THUNK_ENTRY_POINT(886)
THUNK_ENTRY_POINT(887)
THUNK_ENTRY_POINT(888)
THUNK_ENTRY_POINT(889)
THUNK_ENTRY_POINT(890)
THUNK_ENTRY_POINT(891)
THUNK_ENTRY_POINT(892)
THUNK_ENTRY_POINT(893)
THUNK_ENTRY_POINT(894)
THUNK_ENTRY_POINT(895)
THUNK_ENTRY_POINT(896)
THUNK_ENTRY_POINT(897)
THUNK_ENTRY_POINT(898)
THUNK_ENTRY_POINT(899)
THUNK_ENTRY_POINT(900)
THUNK_ENTRY_POINT(901)
THUNK_ENTRY_POINT(902)
THUNK_ENTRY_POINT(903)
THUNK_ENTRY_POINT(904)
THUNK_ENTRY_POINT(905)
THUNK_ENTRY_POINT(906)
THUNK_ENTRY_POINT(907)
THUNK_ENTRY_POINT(908)
THUNK_ENTRY_POINT(909)
THUNK_ENTRY_POINT(910)
THUNK_ENTRY_POINT(911)
THUNK_ENTRY_POINT(912)
THUNK_ENTRY_POINT(913)
THUNK_ENTRY_POINT(914)
THUNK_ENTRY_POINT(915)
THUNK_ENTRY_POINT(916)
THUNK_ENTRY_POINT(917)
THUNK_ENTRY_POINT(918)
THUNK_ENTRY_POINT(919)
THUNK_ENTRY_POINT(920)
THUNK_ENTRY_POINT(921)
THUNK_ENTRY_POINT(922)
THUNK_ENTRY_POINT(923)
THUNK_ENTRY_POINT(924)
THUNK_ENTRY_POINT(925)
THUNK_ENTRY_POINT(926)
THUNK_ENTRY_POINT(927)
THUNK_ENTRY_POINT(928)
THUNK_ENTRY_POINT(929)
THUNK_ENTRY_POINT(930)
THUNK_ENTRY_POINT(931)
THUNK_ENTRY_POINT(932)
THUNK_ENTRY_POINT(933)
THUNK_ENTRY_POINT(934)
THUNK_ENTRY_POINT(935)
THUNK_ENTRY_POINT(936)
THUNK_ENTRY_POINT(937)
THUNK_ENTRY_POINT(938)
THUNK_ENTRY_POINT(939)
THUNK_ENTRY_POINT(940)
THUNK_ENTRY_POINT(941)
THUNK_ENTRY_POINT(942)
THUNK_ENTRY_POINT(943)
THUNK_ENTRY_POINT(944)
THUNK_ENTRY_POINT(945)
THUNK_ENTRY_POINT(946)
THUNK_ENTRY_POINT(947)
THUNK_ENTRY_POINT(948)
THUNK_ENTRY_POINT(949)
THUNK_ENTRY_POINT(950)
THUNK_ENTRY_POINT(951)
THUNK_ENTRY_POINT(952)
THUNK_ENTRY_POINT(953)
THUNK_ENTRY_POINT(954)
THUNK_ENTRY_POINT(955)
THUNK_ENTRY_POINT(956)
THUNK_ENTRY_POINT(957)
THUNK_ENTRY_POINT(958)
THUNK_ENTRY_POINT(959)
THUNK_ENTRY_POINT(960)
THUNK_ENTRY_POINT(961)
THUNK_ENTRY_POINT(962)
THUNK_ENTRY_POINT(963)
THUNK_ENTRY_POINT(964)
THUNK_ENTRY_POINT(965)
THUNK_ENTRY_POINT(966)
THUNK_ENTRY_POINT(967)
THUNK_ENTRY_POINT(968)
THUNK_ENTRY_POINT(969)
THUNK_ENTRY_POINT(970)
THUNK_ENTRY_POINT(971)
THUNK_ENTRY_POINT(972)
THUNK_ENTRY_POINT(973)
THUNK_ENTRY_POINT(974)
THUNK_ENTRY_POINT(975)
THUNK_ENTRY_POINT(976)
THUNK_ENTRY_POINT(977)
THUNK_ENTRY_POINT(978)
THUNK_ENTRY_POINT(979)
THUNK_ENTRY_POINT(980)
THUNK_ENTRY_POINT(981)
THUNK_ENTRY_POINT(982)
THUNK_ENTRY_POINT(983)
THUNK_ENTRY_POINT(984)
THUNK_ENTRY_POINT(985)
THUNK_ENTRY_POINT(986)
THUNK_ENTRY_POINT(987)
THUNK_ENTRY_POINT(988)
THUNK_ENTRY_POINT(989)
THUNK_ENTRY_POINT(990)
THUNK_ENTRY_POINT(991)
THUNK_ENTRY_POINT(992)
THUNK_ENTRY_POINT(993)
THUNK_ENTRY_POINT(994)
THUNK_ENTRY_POINT(995)
THUNK_ENTRY_POINT(996)
THUNK_ENTRY_POINT(997)
THUNK_ENTRY_POINT(998)
THUNK_ENTRY_POINT(999)
THUNK_ENTRY_POINT(1000)
THUNK_ENTRY_POINT(1001)
THUNK_ENTRY_POINT(1002)
THUNK_ENTRY_POINT(1003)
THUNK_ENTRY_POINT(1004)
THUNK_ENTRY_POINT(1005)
THUNK_ENTRY_POINT(1006)
THUNK_ENTRY_POINT(1007)
THUNK_ENTRY_POINT(1008)
THUNK_ENTRY_POINT(1009)
THUNK_ENTRY_POINT(1010)
THUNK_ENTRY_POINT(1011)
THUNK_ENTRY_POINT(1012)
THUNK_ENTRY_POINT(1013)
THUNK_ENTRY_POINT(1014)
THUNK_ENTRY_POINT(1015)
THUNK_ENTRY_POINT(1016)
THUNK_ENTRY_POINT(1017)
THUNK_ENTRY_POINT(1018)
THUNK_ENTRY_POINT(1019)
THUNK_ENTRY_POINT(1020)
THUNK_ENTRY_POINT(1021)
THUNK_ENTRY_POINT(1022)
THUNK_ENTRY_POINT(1023)

static const void* const vtbl[] =
					{	Delegator_QueryInterface,
						Delegator_AddRef,
						Delegator_Release,
						tep_3,
						tep_4,
						tep_5,
						tep_6,
						tep_7,
						tep_8,
						tep_9,
						tep_10,
						tep_11,
						tep_12,
						tep_13,
						tep_14,
						tep_15,
						tep_16,
						tep_17,
						tep_18,
						tep_19,
						tep_20,
						tep_21,
						tep_22,
						tep_23,
						tep_24,
						tep_25,
						tep_26,
						tep_27,
						tep_28,
						tep_29,
						tep_30,
						tep_31,
						tep_32,
						tep_33,
						tep_34,
						tep_35,
						tep_36,
						tep_37,
						tep_38,
						tep_39,
						tep_40,
						tep_41,
						tep_42,
						tep_43,
						tep_44,
						tep_45,
						tep_46,
						tep_47,
						tep_48,
						tep_49,
						tep_50,
						tep_51,
						tep_52,
						tep_53,
						tep_54,
						tep_55,
						tep_56,
						tep_57,
						tep_58,
						tep_59,
						tep_60,
						tep_61,
						tep_62,
						tep_63,
						tep_64,
						tep_65,
						tep_66,
						tep_67,
						tep_68,
						tep_69,
						tep_70,
						tep_71,
						tep_72,
						tep_73,
						tep_74,
						tep_75,
						tep_76,
						tep_77,
						tep_78,
						tep_79,
						tep_80,
						tep_81,
						tep_82,
						tep_83,
						tep_84,
						tep_85,
						tep_86,
						tep_87,
						tep_88,
						tep_89,
						tep_90,
						tep_91,
						tep_92,
						tep_93,
						tep_94,
						tep_95,
						tep_96,
						tep_97,
						tep_98,
						tep_99,
						tep_100,
						tep_101,
						tep_102,
						tep_103,
						tep_104,
						tep_105,
						tep_106,
						tep_107,
						tep_108,
						tep_109,
						tep_110,
						tep_111,
						tep_112,
						tep_113,
						tep_114,
						tep_115,
						tep_116,
						tep_117,
						tep_118,
						tep_119,
						tep_120,
						tep_121,
						tep_122,
						tep_123,
						tep_124,
						tep_125,
						tep_126,
						tep_127,
						tep_128,
						tep_129,
						tep_130,
						tep_131,
						tep_132,
						tep_133,
						tep_134,
						tep_135,
						tep_136,
						tep_137,
						tep_138,
						tep_139,
						tep_140,
						tep_141,
						tep_142,
						tep_143,
						tep_144,
						tep_145,
						tep_146,
						tep_147,
						tep_148,
						tep_149,
						tep_150,
						tep_151,
						tep_152,
						tep_153,
						tep_154,
						tep_155,
						tep_156,
						tep_157,
						tep_158,
						tep_159,
						tep_160,
						tep_161,
						tep_162,
						tep_163,
						tep_164,
						tep_165,
						tep_166,
						tep_167,
						tep_168,
						tep_169,
						tep_170,
						tep_171,
						tep_172,
						tep_173,
						tep_174,
						tep_175,
						tep_176,
						tep_177,
						tep_178,
						tep_179,
						tep_180,
						tep_181,
						tep_182,
						tep_183,
						tep_184,
						tep_185,
						tep_186,
						tep_187,
						tep_188,
						tep_189,
						tep_190,
						tep_191,
						tep_192,
						tep_193,
						tep_194,
						tep_195,
						tep_196,
						tep_197,
						tep_198,
						tep_199,
						tep_200,
						tep_201,
						tep_202,
						tep_203,
						tep_204,
						tep_205,
						tep_206,
						tep_207,
						tep_208,
						tep_209,
						tep_210,
						tep_211,
						tep_212,
						tep_213,
						tep_214,
						tep_215,
						tep_216,
						tep_217,
						tep_218,
						tep_219,
						tep_220,
						tep_221,
						tep_222,
						tep_223,
						tep_224,
						tep_225,
						tep_226,
						tep_227,
						tep_228,
						tep_229,
						tep_230,
						tep_231,
						tep_232,
						tep_233,
						tep_234,
						tep_235,
						tep_236,
						tep_237,
						tep_238,
						tep_239,
						tep_240,
						tep_241,
						tep_242,
						tep_243,
						tep_244,
						tep_245,
						tep_246,
						tep_247,
						tep_248,
						tep_249,
						tep_250,
						tep_251,
						tep_252,
						tep_253,
						tep_254,
						tep_255,
						tep_256,
						tep_257,
						tep_258,
						tep_259,
						tep_260,
						tep_261,
						tep_262,
						tep_263,
						tep_264,
						tep_265,
						tep_266,
						tep_267,
						tep_268,
						tep_269,
						tep_270,
						tep_271,
						tep_272,
						tep_273,
						tep_274,
						tep_275,
						tep_276,
						tep_277,
						tep_278,
						tep_279,
						tep_280,
						tep_281,
						tep_282,
						tep_283,
						tep_284,
						tep_285,
						tep_286,
						tep_287,
						tep_288,
						tep_289,
						tep_290,
						tep_291,
						tep_292,
						tep_293,
						tep_294,
						tep_295,
						tep_296,
						tep_297,
						tep_298,
						tep_299,
						tep_300,
						tep_301,
						tep_302,
						tep_303,
						tep_304,
						tep_305,
						tep_306,
						tep_307,
						tep_308,
						tep_309,
						tep_310,
						tep_311,
						tep_312,
						tep_313,
						tep_314,
						tep_315,
						tep_316,
						tep_317,
						tep_318,
						tep_319,
						tep_320,
						tep_321,
						tep_322,
						tep_323,
						tep_324,
						tep_325,
						tep_326,
						tep_327,
						tep_328,
						tep_329,
						tep_330,
						tep_331,
						tep_332,
						tep_333,
						tep_334,
						tep_335,
						tep_336,
						tep_337,
						tep_338,
						tep_339,
						tep_340,
						tep_341,
						tep_342,
						tep_343,
						tep_344,
						tep_345,
						tep_346,
						tep_347,
						tep_348,
						tep_349,
						tep_350,
						tep_351,
						tep_352,
						tep_353,
						tep_354,
						tep_355,
						tep_356,
						tep_357,
						tep_358,
						tep_359,
						tep_360,
						tep_361,
						tep_362,
						tep_363,
						tep_364,
						tep_365,
						tep_366,
						tep_367,
						tep_368,
						tep_369,
						tep_370,
						tep_371,
						tep_372,
						tep_373,
						tep_374,
						tep_375,
						tep_376,
						tep_377,
						tep_378,
						tep_379,
						tep_380,
						tep_381,
						tep_382,
						tep_383,
						tep_384,
						tep_385,
						tep_386,
						tep_387,
						tep_388,
						tep_389,
						tep_390,
						tep_391,
						tep_392,
						tep_393,
						tep_394,
						tep_395,
						tep_396,
						tep_397,
						tep_398,
						tep_399,
						tep_400,
						tep_401,
						tep_402,
						tep_403,
						tep_404,
						tep_405,
						tep_406,
						tep_407,
						tep_408,
						tep_409,
						tep_410,
						tep_411,
						tep_412,
						tep_413,
						tep_414,
						tep_415,
						tep_416,
						tep_417,
						tep_418,
						tep_419,
						tep_420,
						tep_421,
						tep_422,
						tep_423,
						tep_424,
						tep_425,
						tep_426,
						tep_427,
						tep_428,
						tep_429,
						tep_430,
						tep_431,
						tep_432,
						tep_433,
						tep_434,
						tep_435,
						tep_436,
						tep_437,
						tep_438,
						tep_439,
						tep_440,
						tep_441,
						tep_442,
						tep_443,
						tep_444,
						tep_445,
						tep_446,
						tep_447,
						tep_448,
						tep_449,
						tep_450,
						tep_451,
						tep_452,
						tep_453,
						tep_454,
						tep_455,
						tep_456,
						tep_457,
						tep_458,
						tep_459,
						tep_460,
						tep_461,
						tep_462,
						tep_463,
						tep_464,
						tep_465,
						tep_466,
						tep_467,
						tep_468,
						tep_469,
						tep_470,
						tep_471,
						tep_472,
						tep_473,
						tep_474,
						tep_475,
						tep_476,
						tep_477,
						tep_478,
						tep_479,
						tep_480,
						tep_481,
						tep_482,
						tep_483,
						tep_484,
						tep_485,
						tep_486,
						tep_487,
						tep_488,
						tep_489,
						tep_490,
						tep_491,
						tep_492,
						tep_493,
						tep_494,
						tep_495,
						tep_496,
						tep_497,
						tep_498,
						tep_499,
						tep_500,
						tep_501,
						tep_502,
						tep_503,
						tep_504,
						tep_505,
						tep_506,
						tep_507,
						tep_508,
						tep_509,
						tep_510,
						tep_511,
						tep_512,
						tep_513,
						tep_514,
						tep_515,
						tep_516,
						tep_517,
						tep_518,
						tep_519,
						tep_520,
						tep_521,
						tep_522,
						tep_523,
						tep_524,
						tep_525,
						tep_526,
						tep_527,
						tep_528,
						tep_529,
						tep_530,
						tep_531,
						tep_532,
						tep_533,
						tep_534,
						tep_535,
						tep_536,
						tep_537,
						tep_538,
						tep_539,
						tep_540,
						tep_541,
						tep_542,
						tep_543,
						tep_544,
						tep_545,
						tep_546,
						tep_547,
						tep_548,
						tep_549,
						tep_550,
						tep_551,
						tep_552,
						tep_553,
						tep_554,
						tep_555,
						tep_556,
						tep_557,
						tep_558,
						tep_559,
						tep_560,
						tep_561,
						tep_562,
						tep_563,
						tep_564,
						tep_565,
						tep_566,
						tep_567,
						tep_568,
						tep_569,
						tep_570,
						tep_571,
						tep_572,
						tep_573,
						tep_574,
						tep_575,
						tep_576,
						tep_577,
						tep_578,
						tep_579,
						tep_580,
						tep_581,
						tep_582,
						tep_583,
						tep_584,
						tep_585,
						tep_586,
						tep_587,
						tep_588,
						tep_589,
						tep_590,
						tep_591,
						tep_592,
						tep_593,
						tep_594,
						tep_595,
						tep_596,
						tep_597,
						tep_598,
						tep_599,
						tep_600,
						tep_601,
						tep_602,
						tep_603,
						tep_604,
						tep_605,
						tep_606,
						tep_607,
						tep_608,
						tep_609,
						tep_610,
						tep_611,
						tep_612,
						tep_613,
						tep_614,
						tep_615,
						tep_616,
						tep_617,
						tep_618,
						tep_619,
						tep_620,
						tep_621,
						tep_622,
						tep_623,
						tep_624,
						tep_625,
						tep_626,
						tep_627,
						tep_628,
						tep_629,
						tep_630,
						tep_631,
						tep_632,
						tep_633,
						tep_634,
						tep_635,
						tep_636,
						tep_637,
						tep_638,
						tep_639,
						tep_640,
						tep_641,
						tep_642,
						tep_643,
						tep_644,
						tep_645,
						tep_646,
						tep_647,
						tep_648,
						tep_649,
						tep_650,
						tep_651,
						tep_652,
						tep_653,
						tep_654,
						tep_655,
						tep_656,
						tep_657,
						tep_658,
						tep_659,
						tep_660,
						tep_661,
						tep_662,
						tep_663,
						tep_664,
						tep_665,
						tep_666,
						tep_667,
						tep_668,
						tep_669,
						tep_670,
						tep_671,
						tep_672,
						tep_673,
						tep_674,
						tep_675,
						tep_676,
						tep_677,
						tep_678,
						tep_679,
						tep_680,
						tep_681,
						tep_682,
						tep_683,
						tep_684,
						tep_685,
						tep_686,
						tep_687,
						tep_688,
						tep_689,
						tep_690,
						tep_691,
						tep_692,
						tep_693,
						tep_694,
						tep_695,
						tep_696,
						tep_697,
						tep_698,
						tep_699,
						tep_700,
						tep_701,
						tep_702,
						tep_703,
						tep_704,
						tep_705,
						tep_706,
						tep_707,
						tep_708,
						tep_709,
						tep_710,
						tep_711,
						tep_712,
						tep_713,
						tep_714,
						tep_715,
						tep_716,
						tep_717,
						tep_718,
						tep_719,
						tep_720,
						tep_721,
						tep_722,
						tep_723,
						tep_724,
						tep_725,
						tep_726,
						tep_727,
						tep_728,
						tep_729,
						tep_730,
						tep_731,
						tep_732,
						tep_733,
						tep_734,
						tep_735,
						tep_736,
						tep_737,
						tep_738,
						tep_739,
						tep_740,
						tep_741,
						tep_742,
						tep_743,
						tep_744,
						tep_745,
						tep_746,
						tep_747,
						tep_748,
						tep_749,
						tep_750,
						tep_751,
						tep_752,
						tep_753,
						tep_754,
						tep_755,
						tep_756,
						tep_757,
						tep_758,
						tep_759,
						tep_760,
						tep_761,
						tep_762,
						tep_763,
						tep_764,
						tep_765,
						tep_766,
						tep_767,
						tep_768,
						tep_769,
						tep_770,
						tep_771,
						tep_772,
						tep_773,
						tep_774,
						tep_775,
						tep_776,
						tep_777,
						tep_778,
						tep_779,
						tep_780,
						tep_781,
						tep_782,
						tep_783,
						tep_784,
						tep_785,
						tep_786,
						tep_787,
						tep_788,
						tep_789,
						tep_790,
						tep_791,
						tep_792,
						tep_793,
						tep_794,
						tep_795,
						tep_796,
						tep_797,
						tep_798,
						tep_799,
						tep_800,
						tep_801,
						tep_802,
						tep_803,
						tep_804,
						tep_805,
						tep_806,
						tep_807,
						tep_808,
						tep_809,
						tep_810,
						tep_811,
						tep_812,
						tep_813,
						tep_814,
						tep_815,
						tep_816,
						tep_817,
						tep_818,
						tep_819,
						tep_820,
						tep_821,
						tep_822,
						tep_823,
						tep_824,
						tep_825,
						tep_826,
						tep_827,
						tep_828,
						tep_829,
						tep_830,
						tep_831,
						tep_832,
						tep_833,
						tep_834,
						tep_835,
						tep_836,
						tep_837,
						tep_838,
						tep_839,
						tep_840,
						tep_841,
						tep_842,
						tep_843,
						tep_844,
						tep_845,
						tep_846,
						tep_847,
						tep_848,
						tep_849,
						tep_850,
						tep_851,
						tep_852,
						tep_853,
						tep_854,
						tep_855,
						tep_856,
						tep_857,
						tep_858,
						tep_859,
						tep_860,
						tep_861,
						tep_862,
						tep_863,
						tep_864,
						tep_865,
						tep_866,
						tep_867,
						tep_868,
						tep_869,
						tep_870,
						tep_871,
						tep_872,
						tep_873,
						tep_874,
						tep_875,
						tep_876,
						tep_877,
						tep_878,
						tep_879,
						tep_880,
						tep_881,
						tep_882,
						tep_883,
						tep_884,
						tep_885,
						tep_886,
						tep_887,
						tep_888,
						tep_889,
						tep_890,
						tep_891,
						tep_892,
						tep_893,
						tep_894,
						tep_895,
						tep_896,
						tep_897,
						tep_898,
						tep_899,
						tep_900,
						tep_901,
						tep_902,
						tep_903,
						tep_904,
						tep_905,
						tep_906,
						tep_907,
						tep_908,
						tep_909,
						tep_910,
						tep_911,
						tep_912,
						tep_913,
						tep_914,
						tep_915,
						tep_916,
						tep_917,
						tep_918,
						tep_919,
						tep_920,
						tep_921,
						tep_922,
						tep_923,
						tep_924,
						tep_925,
						tep_926,
						tep_927,
						tep_928,
						tep_929,
						tep_930,
						tep_931,
						tep_932,
						tep_933,
						tep_934,
						tep_935,
						tep_936,
						tep_937,
						tep_938,
						tep_939,
						tep_940,
						tep_941,
						tep_942,
						tep_943,
						tep_944,
						tep_945,
						tep_946,
						tep_947,
						tep_948,
						tep_949,
						tep_950,
						tep_951,
						tep_952,
						tep_953,
						tep_954,
						tep_955,
						tep_956,
						tep_957,
						tep_958,
						tep_959,
						tep_960,
						tep_961,
						tep_962,
						tep_963,
						tep_964,
						tep_965,
						tep_966,
						tep_967,
						tep_968,
						tep_969,
						tep_970,
						tep_971,
						tep_972,
						tep_973,
						tep_974,
						tep_975,
						tep_976,
						tep_977,
						tep_978,
						tep_979,
						tep_980,
						tep_981,
						tep_982,
						tep_983,
						tep_984,
						tep_985,
						tep_986,
						tep_987,
						tep_988,
						tep_989,
						tep_990,
						tep_991,
						tep_992,
						tep_993,
						tep_994,
						tep_995,
						tep_996,
						tep_997,
						tep_998,
						tep_999,
						tep_1000,
						tep_1001,
						tep_1002,
						tep_1003,
						tep_1004,
						tep_1005,
						tep_1006,
						tep_1007,
						tep_1008,
						tep_1009,
						tep_1010,
						tep_1011,
						tep_1012,
						tep_1013,
						tep_1014,
						tep_1015,
						tep_1016,
						tep_1017,
						tep_1018,
						tep_1019,
						tep_1020,
						tep_1021,
						tep_1022,
						tep_1023 };

const void* const Delegator::s_vptr = vtbl;
